# SPDX-License-Identifier: MIT

"""Models for Znuny API calls"""

from enum import IntEnum, StrEnum
from logging import getLogger
from typing import Annotated, Any

from pydantic import (
    BaseModel,
    BeforeValidator,
    ConfigDict,
    Field,
    model_validator,
)

logger = getLogger(__name__)


def _empty_str_to_none(value: Any) -> Any:
    return None if value == "" else value


class ServiceID(IntEnum):
    """Enum for Service_IDs"""

    FIREWALL = 264
    SHAREPOINT = 112
    VIRTUAL_SERVER = 69
    GROUP_MANAGEMENT = 86
    POOL_RESERVATION = 98
    EXCHANGE_MAIL = 63
    INTRANET = 115


class Boolean(IntEnum):
    """Enum for int values used as bool in Znuny"""

    TRUE = 1
    FALSE = 0


class CommunicationChannel(IntEnum):
    """Enum for article communication channels"""

    EMAIL = 1
    PHONE = 2
    INTERNAL = 3
    CHAT = 4


class QueueID(IntEnum):
    """Enum for queue IDs"""

    CIDS_IT_BETRIEB_WINDOWS = 1047
    CIDS_SERVICEDESK = 1038
    CIDS = 1037
    CIDS_IT_BETRIEB_LINUX = 1046
    INFORMATIONSSICHERHEIT = 124
    CIDS_DIENSTEENTWICKLUNG_SOFTWAREENTWICKLUNG = 1052
    AUTOMAT_SSP = 1074
    CIDS_IT_BETRIEB_NETZ = 1048
    CIDS_IT_BETRIEB_BASISDIENSTE = 1045
    INTRANET = 1024


class TicketType(IntEnum):
    """Enum for ticket types"""

    UNCLASSIFIED = 1
    INCIDENT = 2
    SERVICE_REQUEST = 4
    PROBLEM = 5
    CHANGE_REQUEST = 6
    CONSULTING_SERVICE = 7


class TicketState(IntEnum):
    """Enum ticket states"""

    NEW = 1
    CONFIRM_ANSWER = 2
    IN_PROCESS = 3
    RESOLVED = 4
    REMOVED = 5
    WAITING_FOR_INFORMATION = 6
    CLOSED_WITHOUT_RESOLUTION = 7
    WAITING_FOR_EXTERNAL_SERVICE_PROVIDER = 8
    MERGED = 9
    CONFIRM_SOLUTION = 10
    IN_TEST = 11
    WAITING_FOR_RESUBMISSION = 12


class TicketPriority(IntEnum):
    """Enum for Znuny ticket priorities"""

    URGENT = 5
    HIGH = 4
    NORMAL = 3
    LOW = 1


class SenderType(IntEnum):
    "SenderType for Znuny article"

    AGENT = 1
    CUSTOMER = 3
    SYSTEM = 2


class DynamicFieldName(StrEnum):
    """names of dynamic fields"""

    REASON_FREETEXT = "ReasonFreetext"
    REASON_SELECTOR = "ReasonSelector"  # Anfragegrund


class ServiceName(StrEnum):
    """values of dynamic Category fields"""

    FIREWALL = "Firewall"
    SHAREPOINT = "Kollaborationstools::SharePoint"
    VIRTUAL_SERVER = "Virtualisierung"


class DynamicFieldValueReasonSelector(StrEnum):
    """values of dynamic ReasonSelector fields"""

    PROVISIONING = "04_Provisioning"


class PendingTime(BaseModel):
    """Pending time"""

    diff: Annotated[int, Field(alias="Diff")]


class Ticket(BaseModel):
    """Model for a Znuny ticket"""

    model_config = ConfigDict(populate_by_name=True)

    ticket_id: Annotated[
        int | None,
        Field(alias="TicketID"),
    ] = None
    ticket_number: Annotated[
        str | None,
        Field(alias="TicketNumber"),
    ] = None
    title: Annotated[
        str,
        Field(alias="Title"),
    ]
    queue_name: Annotated[
        str | None,
        Field(alias="Queue"),
    ] = None
    queue_id: Annotated[
        int | None,
        Field(alias="QueueID"),
    ] = None
    state_id: Annotated[
        TicketState,
        Field(alias="StateID"),
    ] = TicketState.NEW
    priority_id: Annotated[
        TicketPriority,
        Field(alias="PriorityID"),
    ] = TicketPriority.NORMAL
    type_id: Annotated[
        TicketType,
        Field(alias="TypeID"),
    ] = TicketType.SERVICE_REQUEST
    customer_user: Annotated[
        str | None,
        Field(alias="CustomerUser"),
    ] = None
    customer_user_id: Annotated[
        str | None,
        Field(alias="CustomerUserID"),
    ] = None
    service: Annotated[
        str | None,
        Field(alias="Service"),
    ] = None
    service_id: Annotated[
        int | None,
        Field(alias="ServiceID"),
        BeforeValidator(_empty_str_to_none),
    ] = None

    @classmethod
    @model_validator(mode="before")
    def check_queue_or_id_given(cls, values: dict[str, Any]) -> None:
        """Check if either Queue or QueueID is given"""
        if values.get("queue_name") is None and values.get("queue_id") is None:
            raise AssertionError("Either 'Queue' or 'QueueID' must be given.")


class TicketUpdate(BaseModel):
    """Model for update to Znuny ticket"""

    model_config = ConfigDict(populate_by_name=True)

    queue_name: Annotated[
        str | None,
        Field(alias="Queue"),
    ] = None
    queue_id: Annotated[
        int | None,
        Field(alias="QueueID"),
    ] = None
    state_id: Annotated[
        TicketState | None,
        Field(alias="StateID"),
    ] = None
    pending_time: Annotated[
        PendingTime | None,
        Field(alias="PendingTime"),
    ] = None

    @classmethod
    @model_validator(mode="before")
    def check_only_one_of_queue_id_or_name_given(
        cls, values: dict[str, Any]
    ) -> None:
        """Check that at most one of queue_name and queue_id is specified"""
        if "queue_name" in values and "queue_id" in values:
            raise ValueError(
                "At most one of QueueID and Queue can be specified"
            )


class Attachment(BaseModel):
    """Model for a Znuny attachment"""

    model_config = ConfigDict(populate_by_name=True)

    content_b64: Annotated[
        str,
        Field(alias="Content"),
    ]
    content_type: Annotated[
        str,
        Field(alias="ContentType"),
    ]
    filename: Annotated[
        str,
        Field(alias="Filename"),
    ]


class Article(BaseModel):
    """Model for a Znuny article"""

    model_config = ConfigDict(populate_by_name=True)

    article_id: Annotated[
        int | None,
        Field(alias="ArticleID"),
    ] = None
    subject: Annotated[
        str,
        Field(alias="Subject"),
    ]
    body: Annotated[
        str,
        Field(alias="Body"),
    ]
    communication_channel_id: Annotated[
        CommunicationChannel,
        Field(alias="CommunicationChannelID"),
    ] = CommunicationChannel.EMAIL
    to: Annotated[
        str,
        Field(alias="To"),
    ]
    cc: Annotated[
        str | None,
        Field(alias="Cc"),
    ] = None
    bcc: Annotated[
        str | None,
        Field(alias="Bcc"),
    ] = None
    content_type: Annotated[
        str,
        Field(alias="ContentType"),
    ] = "text/html; charset=utf-8"
    history_type: Annotated[
        str | None,
        Field(alias="HistoryType"),
    ] = "SystemRequest"
    time_unit: Annotated[
        int,
        Field(alias="TimeUnit"),
    ] = 0
    send_article: Annotated[
        Boolean,
        Field(alias="ArticleSend"),
    ] = Boolean.TRUE
    visible_for_customer: Annotated[
        Boolean,
        Field(alias="IsVisibleForCustomer"),
    ] = Boolean.TRUE
    no_agent_notify: Annotated[
        Boolean,
        Field(alias="NoAgentNotify"),
    ] = Boolean.TRUE
    sender_type: Annotated[
        SenderType,
        Field(alias="SenderTypeID"),
    ] = SenderType.AGENT
    auto_sign: Annotated[
        Boolean | None,
        Field(alias="AutoSign"),
    ] = Boolean.TRUE



# SPDX-License-Identifier: MIT

"""Tests for the Znuny Connection"""

from typing import Any
from unittest import TestCase
from unittest.mock import Mock

from httpx import HTTPError

from .connection import Connection


class TestConnection(TestCase):
    """Simple connection tests by mocking the session"""

    test_ticket_data: dict[str, dict[str, Any]] = {
        "Ticket": {
            "Title": "Testticket",
            "QueueID": 1,
            "CustomerUser": "testuser",
        },
        "Article": {
            "Subject": "Testticket",
            "Body": "Nur ein Testticket.",
            "To": "testuser@tu-dresden.de",
        },
    }

    client_mock = Mock()
    response_mock = Mock()

    def test_create_ticket(self) -> None:
        """create a ticket"""
        self.response_mock.status_code = 200
        self.response_mock.json = Mock(
            return_value={
                "Ticket": {
                    "Title": "Testticket",
                    "TicketID": 123,
                    "QueueID": 1,
                    "CustomerUser": "testuser",
                }
            }
        )
        self.client_mock.post = Mock(return_value=self.response_mock)
        connection = Connection(
            "fake_url", "fake_user", "fake_password", client=self.client_mock
        )
        connection.create_ticket(self.test_ticket_data)
        self.client_mock.post.assert_called_once_with(
            "fake_url/Ticket/", json=self.test_ticket_data
        )

    def test_create_ticket_raises(self) -> None:
        """ticket creation raises an HTTPError"""
        self.response_mock.status_code = 400
        self.response_mock.raise_for_status = Mock(
            side_effect=HTTPError("foo")
        )
        self.client_mock.post = Mock(return_value=self.response_mock)
        connection = Connection(
            "fake_url", "fake_user", "fake_password", client=self.client_mock
        )
        with self.assertRaises(HTTPError):
            connection.create_ticket(self.test_ticket_data)

    def test_update_ticket(self) -> None:
        """update a ticket"""
        self.client_mock.patch = Mock()
        connection = Connection(
            "fake_url", "fake_user", "fake_password", client=self.client_mock
        )
        self.test_ticket_data["Ticket"]["QueueID"] = 2
        connection.update_ticket(123456, self.test_ticket_data)
        self.client_mock.patch.assert_called_once_with(
            "fake_url/Ticket/123456/", json=self.test_ticket_data
        )



# SPDX-License-Identifier: MIT

"""Tests for Znuny models"""

from unittest import TestCase
from unittest.mock import Mock, patch

from typing_extensions import override

from .connection import Connection
from .manager import TicketManager
from .models import (
    Article,
    Boolean,
    CommunicationChannel,
    QueueID,
    SenderType,
    Ticket,
    TicketPriority,
    TicketState,
    TicketType,
    TicketUpdate,
)


class TestModels(TestCase):
    """Testcases for Znuny models"""

    test_title = "Testticket"
    test_body = "Dies ist ein Testticket."
    test_queue_id = QueueID.CIDS_DIENSTEENTWICKLUNG_SOFTWAREENTWICKLUNG
    test_customer = "testcustomer"
    test_to = "testcustomer@tu-dresden.de"

    @override
    def setUp(self) -> None:
        """Set up each test"""
        self.test_ticket = Ticket.model_construct(
            ticket_id=42,
            title=self.test_title,
            queue_id=self.test_queue_id,
            customer_user=self.test_customer,
        )
        self.test_article = Article.model_construct(
            subject=self.test_title, body=self.test_body, to=self.test_to
        )

    def test_create_ticket(self) -> None:
        """Create a ticket"""
        client_mock = Mock()
        response_mock = Mock()
        response_mock.json = Mock(
            return_value={"Ticket": self.test_ticket.model_dump(by_alias=True)}
        )
        client_mock.post = Mock(return_value=response_mock)
        connection_mock = Connection(
            "test_url", "test_user", "test_password", client=client_mock
        )
        manager = TicketManager(connection_mock)
        returned_ticket = manager.create(
            queue_id=self.test_queue_id,
            customer=self.test_customer,
            title=self.test_title,
            body=self.test_body,
            to=self.test_to,
        )
        self.assertIsNotNone(returned_ticket)
        self.assertEqual(self.test_ticket, returned_ticket)

    def test_create_and_move_ticket(self) -> None:
        """Create a ticket"""
        connection_mock = Mock()
        connection_mock.create_ticket = Mock(
            return_value={"Ticket": self.test_ticket.model_dump(by_alias=True)}
        )
        connection_mock.update_ticket = Mock(return_value={})
        manager = TicketManager(connection_mock)
        returned_ticket = manager.create_and_move(
            queue_id=self.test_queue_id,
            customer=self.test_customer,
            title=self.test_title,
            body=self.test_body,
            to=self.test_to,
        )
        self.assertEqual(self.test_ticket, returned_ticket)
        connection_mock.update_ticket.assert_called_once_with(
            42,
            {
                "Ticket": {
                    "QueueID": self.test_queue_id,
                },
            },
        )

    def test_update_ticket(self) -> None:
        """Update a ticket"""
        connection = Mock()
        connection.update_ticket = Mock(return_value={})
        manager = TicketManager(connection)
        ticket = TicketUpdate.model_construct(queue_name="Some::Queue")
        manager.update(
            ticket_id=42,
            ticket=ticket,
        )
        connection.update_ticket.assert_called_once_with(
            42,
            {
                "Ticket": {
                    "Queue": "Some::Queue",
                },
            },
        )

    def test_call_model_with_sender_type(self) -> None:
        """Test the new sender type"""
        client_mock = Mock()
        response_mock = Mock()
        response_mock.json = Mock(
            return_value={"Ticket": self.test_ticket.model_dump(by_alias=True)}
        )
        client_mock.post = Mock(return_value=response_mock)
        connection_mock = Connection(
            "test_url", "test_user", "test_password", client=client_mock
        )
        with patch(
            "zih_clients.znuny.connection.Connection.create_ticket",
        ) as create_ticket_mock:
            create_ticket_mock.return_value = {"Ticket": self.test_ticket}

            manager = TicketManager(connection_mock)
            _ = manager.create(
                queue_id=self.test_queue_id,
                customer=self.test_customer,
                title=self.test_title,
                body=self.test_body,
                to=self.test_to,
                sender_type=SenderType.CUSTOMER,
                CommunicationChannelID=CommunicationChannel.PHONE,
            )

            create_ticket_mock.assert_called_once_with(
                {
                    "Ticket": {
                        "Title": "Testticket",
                        "QueueID": QueueID.CIDS_DIENSTEENTWICKLUNG_SOFTWAREENTWICKLUNG,
                        "StateID": TicketState.NEW,
                        "PriorityID": TicketPriority.NORMAL,
                        "TypeID": TicketType.SERVICE_REQUEST,
                        "CustomerUser": "testcustomer",
                    },
                    "Article": {
                        "Subject": "Testticket",
                        "Body": "Dies ist ein Testticket.",
                        "CommunicationChannelID": CommunicationChannel.PHONE,
                        "To": "testcustomer@tu-dresden.de",
                        "ContentType": "text/html; charset=utf-8",
                        "HistoryType": "SystemRequest",
                        "TimeUnit": 0,
                        "ArticleSend": Boolean.TRUE,
                        "IsVisibleForCustomer": Boolean.TRUE,
                        "NoAgentNotify": Boolean.TRUE,
                        "SenderTypeID": SenderType.CUSTOMER,
                        "AutoSign": Boolean.TRUE,
                    },
                }
            )



# SPDX-License-Identifier: MIT

from logging import getLogger
from typing import Any, Union

from .connection import Connection
from .models import (
    Article,
    Attachment,
    DynamicFieldName,
    QueueID,
    ServiceID,
    ServiceName,
    Ticket,
    TicketUpdate,
)
from .types import ZnunyException

logger = getLogger(__name__)


def _serialize(
    attachment: Union["Attachment", list["Attachment"], None] = None,
) -> dict[str, Any] | list[dict[str, Any]] | None:
    match attachment:
        case Attachment():
            return attachment.model_dump(by_alias=True)
        case list():
            return [a.model_dump(by_alias=True) for a in attachment]
        case _:
            return None


class TicketManager:
    """Manager for ticket objects"""

    def __init__(
        self,
        connection: Connection,
        *,
        temporary_queue_id: QueueID = QueueID.AUTOMAT_SSP,
        override_queue_id: QueueID | int | None = None,
        override_queue_name: str | None = None,
    ):
        """
        initialize ticket manager

        Using `override_queue_id` and `override_queue_name` all
        tickets created or updated can be redirected to the given
        queue. This is intended to be used for development and testing
        purposes only.

        Args:
            connection: low-level connection to use
            override_queue_id: id of queue to redirect all tickets to
            override_queue_name: name of queue to redirect all tickets to
        """

        self._connection = connection
        self._temporary_queue_id = temporary_queue_id
        self._override_queue_id = override_queue_id
        self._override_queue_name = override_queue_name

    def _override_queue(
        self, queue_id: QueueID | int | None, queue_name: str | None
    ) -> tuple[QueueID | int | None, str | None]:
        if queue_id is None and queue_name is None:
            return None, None
        if self._override_queue_id is not None:
            queue_id = self._override_queue_id
            queue_name = None
        if self._override_queue_name is not None:
            queue_id = None
            queue_name = self._override_queue_name
        return queue_id, queue_name

    # pylint: disable=too-many-arguments,too-many-locals
    def create(
        self,
        *,
        queue_id: QueueID | int | None = None,
        queue_name: str | None = None,
        customer: str,
        title: str,
        body: str,
        to: str,  # pylint: disable=invalid-name
        dynamic_fields: dict[DynamicFieldName | str, Any] | None = None,
        attachment: Union["Attachment", list["Attachment"], None] = None,
        service_id: ServiceID | None = None,
        service_name: ServiceName | None = None,
        **kwargs: Any,
    ) -> "Ticket":
        """Return a newly created ticket

        Exactly one of queue_id and queue_name must be specified.

        Args:
            queue_id: QueueID to position ticket into
            queue_name: Name of queue to position ticket into
            customer: ZIH-Login the ticket will be linked to
            title: title of the new ticket and first article
            body: content of the first ticket article
            to: email address the first article will be sent to
            dynamic_fields: dictionary of dynamic fields to set for the ticket

        Returns:
            the newly created :class:`Ticket`

        Raises:
            :class:`ZnunyException` if there was an error creating the ticket

        """
        if queue_id is None and queue_name is None:
            raise ValueError("Must specify one of queue_id or queue_name")
        if queue_id is not None and queue_name is not None:
            raise ValueError("Cannot specify both queue_id and queue_name")
        if service_id is not None and service_name is not None:
            raise ValueError("Cannot specify both service_id and service_name")

        queue_id, queue_name = self._override_queue(queue_id, queue_name)

        if queue_id is not None:
            kwargs["QueueID"] = queue_id
        if queue_name is not None:
            kwargs["Queue"] = queue_name

        if service_id is not None:
            kwargs["ServiceID"] = service_id
        if service_name is not None:
            kwargs["Service"] = service_name

        ticket = Ticket.model_construct(
            title=title, customer_user=customer, **kwargs
        )
        article = Article.model_construct(
            subject=title, body=body, to=to, **kwargs
        )
        params: dict[str, Any] = {
            "Ticket": ticket.model_dump(by_alias=True, exclude_none=True),
            "Article": article.model_dump(by_alias=True, exclude_none=True),
        }
        if dynamic_fields:
            params["DynamicField"] = [
                {"Name": key, "Value": value}
                for key, value in dynamic_fields.items()
            ]
        if raw_attachment := _serialize(attachment):
            params["Attachment"] = raw_attachment
        response_dict = self._connection.create_ticket(params)
        if "Error" in response_dict:
            error = response_dict["Error"]
            logger.error("Could not create ticket: %s", error["ErrorMessage"])
            raise ZnunyException(error["ErrorMessage"])
        return Ticket.model_validate(response_dict["Ticket"])

    def update(
        self,
        ticket_id: int,
        *,
        ticket: "TicketUpdate | None" = None,
        article: "Article | None" = None,
        attachment: Union["Attachment", list["Attachment"], None] = None,
    ) -> None:
        """Update existing ticket"""
        kwargs: dict[str, Any] = {}
        if ticket is None and article is None:
            raise ValueError("At least one of ticket and article is required")
        if ticket is not None:
            ticket.queue_id, ticket.queue_name = self._override_queue(
                ticket.queue_id, ticket.queue_name
            )
            kwargs["Ticket"] = ticket.model_dump(
                by_alias=True, exclude_none=True
            )
        if article is not None:
            kwargs["Article"] = article.model_dump(
                by_alias=True, exclude_none=True
            )
        if raw_attachment := _serialize(attachment):
            kwargs["Attachment"] = raw_attachment
        response = self._connection.update_ticket(ticket_id, kwargs)
        if (error := response.get("Error")) is not None:
            raise ZnunyException(error["ErrorMessage"])

    def create_and_move(
        self,
        *,
        queue_id: QueueID | int | None = None,
        queue_name: str | None = None,
        customer: str,
        title: str,
        body: str,
        to: str,  # pylint: disable=invalid-name
        dynamic_fields: dict[DynamicFieldName | str, Any] | None = None,
        service_id: ServiceID | None = None,
        service_name: ServiceName | None = None,
        **kwargs: Any,
    ) -> "Ticket":
        """Return a newly created ticket that is created add `ZIH::Intern::Self Service Portal::Outgoing`
        and than moved to given queue.

        Exactly one of queue_id and queue_name must be specified.

        Args:
            queue_id: QueueID to position ticket into
            queue_name: Name of queue to position ticket into
            customer: ZIH-Login the ticket will be linked to
            title: title of the new ticket and first article
            body: content of the first ticket article
            to: email address the first article will be sent to
            dynamic_fields: dictionary of dynamic fields to set for the ticket

        Returns:
            the newly created :class:`Ticket`

        Raises:
            :class:`ZnunyException` if there was an error creating the ticket

        """
        if queue_id is None and queue_name is None:
            raise ValueError("Must specify one of queue_id or queue_name")
        if queue_id is not None and queue_name is not None:
            raise ValueError("Cannot specify both queue_id and queue_name")
        if service_id is not None and service_name is not None:
            raise ValueError("Cannot specify both service_id and service_name")

        queue_id, queue_name = self._override_queue(queue_id, queue_name)

        if service_id is not None:
            kwargs["ServiceID"] = service_id
        if service_name is not None:
            kwargs["Service"] = service_name

        ticket = self.create(
            queue_id=self._temporary_queue_id,
            customer=customer,
            title=title,
            body=body,
            to=to,
            dynamic_fields=dynamic_fields,
            **kwargs,
        )
        assert isinstance(ticket.ticket_id, int)
        if queue_id is not None:
            ticket_update = TicketUpdate.model_construct(queue_id=queue_id)
        if queue_name is not None:
            ticket_update = TicketUpdate.model_construct(queue_name=queue_name)
        self.update(ticket_id=ticket.ticket_id, ticket=ticket_update)
        return ticket

    def get_ticket_agent_url(self, ticket_id: int) -> str:
        """get URL to the agent view of the given ticket"""

        return (
            "https://tickets.tu-dresden.de/otrs/index.pl?"
            f"Action=AgentTicketZoom;TicketID={ticket_id}"
        )

    def get_ticket_customer_url(self, ticket_number: str) -> str:
        """get URL to the customer view of the given ticket"""

        return (
            "https://tickets.tu-dresden.de/otrs/customer.pl?"
            f"Action=CustomerTicketZoom;TicketNumber={ticket_number}"
        )
